#Kindly borrowed from UnSloth's implementation
import triton
import triton.language as tl
import torch
from .utils import calculate_settings


@triton.jit
def _rms_layernorm_forward(
    Y, Y_row_stride,
    X, X_row_stride,
    W, W_row_stride,
    r, r_row_stride,
    n_cols, eps,
    BLOCK_SIZE : tl.constexpr
):
    """
        Fast RMS Layernorm kernel
        Inspiration from a Triton tutorial:
        https://triton-lang.org/main/getting-started/tutorials/05-layer-norm.html
    """
    row_idx = tl.program_id(0)
    col_offsets = tl.arange(0, BLOCK_SIZE)
    mask = col_offsets < n_cols

    Y += row_idx * Y_row_stride
    X += row_idx * X_row_stride
    r += row_idx * r_row_stride

    X_row = tl.load(X + col_offsets, mask = mask, other = 0).to(tl.float32)
    W_row = tl.load(W + col_offsets, mask = mask, other = 0)#.to(tl.float32)

    row_var = tl.sum(X_row * X_row, axis = 0) / n_cols
    inv_var = tl.math.rsqrt(row_var + eps)
    tl.store(r, inv_var)
    normed = X_row * inv_var
    normed = normed.to(W_row.dtype) # Exact copy from HF
    output = normed * W_row
    tl.store(Y + col_offsets, output, mask = mask)
pass


@triton.heuristics({"GEMMA": lambda args: args["GEMMA"],})
@triton.jit
def _rms_layernorm_backward(
    dY, dY_row_stride,
    X,   X_row_stride,
    W,   W_row_stride,
    r,   r_row_stride,
    dW, dW_row_stride,
    n_cols, eps,
    GEMMA      : tl.constexpr,
    BLOCK_SIZE : tl.constexpr,
):
    """
        Fast RMS Layernorm kernel for the backward pass
        Inspiration from a Triton tutorial:
        https://triton-lang.org/main/getting-started/tutorials/05-layer-norm.html
    """
    row_idx = tl.program_id(0)
    col_offsets = tl.arange(0, BLOCK_SIZE)
    mask = col_offsets < n_cols

    dY += row_idx * dY_row_stride
    X  += row_idx *  X_row_stride
    r  += row_idx *  r_row_stride

    dY_row = tl.load(dY + col_offsets, mask = mask, other = 0).to(tl.float32)
    X_row  = tl.load(X  + col_offsets, mask = mask, other = 0).to(tl.float32)
    W_row  = tl.load(W  + col_offsets, mask = mask, other = 0).to(tl.float32)

    # Get saved row variance
    inv_var = tl.load(r).to(tl.float32)
    normed = X_row * inv_var

    if GEMMA: dY_W = dY_row * (W_row + 1.0)
    else:     dY_W = dY_row * W_row

    rowsum_dY_normed = tl.sum(dY_W * normed, axis = 0)
    output = inv_var/n_cols * (n_cols*dY_W - normed*rowsum_dY_normed)
    tl.store(dY + col_offsets, output, mask = mask)
pass


@triton.jit
def _gemma_rms_layernorm_forward(
    Y, Y_row_stride,
    X, X_row_stride,
    W, W_row_stride,
    r, r_row_stride,
    n_cols, eps,
    BLOCK_SIZE : tl.constexpr,
):
    # Copies https://github.com/google-deepmind/gemma/blob/main/gemma/layers.py#L31
    # and https://github.com/keras-team/keras-nlp/blob/v0.8.2/keras_nlp/models/gemma/rms_normalization.py#L33
    # exactly. Essentially all in float32!
    row_idx = tl.program_id(0)
    col_offsets = tl.arange(0, BLOCK_SIZE)
    mask = col_offsets < n_cols

    Y += row_idx * Y_row_stride
    X += row_idx * X_row_stride
    r += row_idx * r_row_stride

    X_row = tl.load(X + col_offsets, mask = mask, other = 0).to(tl.float32)
    W_row = tl.load(W + col_offsets, mask = mask, other = 0).to(tl.float32)

    row_var = tl.sum(X_row * X_row, axis = 0) / n_cols
    inv_var = 1.0 / tl.sqrt(row_var + eps) # Must be 1/sqrt to match Deepmind's impl
    tl.store(r, inv_var)
    normed = X_row * inv_var
    output = normed * (W_row + 1.0)

    tl.store(Y + col_offsets, output, mask = mask)
pass


class Fast_RMS_Layernorm(torch.autograd.Function):
    @staticmethod
    def forward(ctx, X, W, eps, gemma = False):
        shape = X.shape
        dim = shape[-1]
        X = X.view(-1, dim)
        n_rows, n_cols = X.shape
        BLOCK_SIZE, num_warps = calculate_settings(n_cols)

        Y = torch.empty((n_rows, n_cols), dtype = X.dtype, device = "cuda")
        r = torch.empty(n_rows, dtype = torch.float32, device = "cuda")

        fx = _gemma_rms_layernorm_forward if gemma else _rms_layernorm_forward
        fx[(n_rows,)](
            Y, Y.stride(0),
            X, X.stride(0),
            W, W.stride(0),
            r, r.stride(0),
            n_cols, eps,
            BLOCK_SIZE = BLOCK_SIZE,
            num_warps  = num_warps,
        )
        ctx.eps = eps
        ctx.BLOCK_SIZE = BLOCK_SIZE
        ctx.num_warps  = num_warps
        ctx.GEMMA = gemma
        ctx.save_for_backward(X, W, r)
        return Y.view(*shape)
    pass

    @staticmethod
    def backward(ctx, dY):
        shape = dY.shape
        dim = shape[-1]
        dY = dY.view(-1, dim)
        X, W, r = ctx.saved_tensors
        n_rows, n_cols = dY.shape
        dW = X

        _rms_layernorm_backward[(n_rows,)](
            dY, dY.stride(0),
            X,  X .stride(0),
            W,  W .stride(0),
            r,  r .stride(0),
            dW, dW.stride(0),
            n_cols, ctx.eps,
            GEMMA      = ctx.GEMMA,
            BLOCK_SIZE = ctx.BLOCK_SIZE,
            num_warps  = ctx.num_warps,
        )
        dX = dY.view(*shape)
        return dX, None, None, None
    pass
pass

def fast_rms_layernorm(weight, X, eps, gemma = False):
    out = Fast_RMS_Layernorm.apply(X, weight, eps, gemma)
    return out
pass


# Inspired by the amazing work on DeltaBit https://github.com/FasterDecoding/BitDelta/tree/main and AutoGPTQ https://github.com/AutoGPTQ/AutoGPTQ
import torch
import triton
import triton.language as tl
from ..utils import custom_autotune
@custom_autotune.autotune(
        configs=[
            triton.Config({"BLOCK_SIZE_M": 16, "BLOCK_SIZE_N": 16, "BLOCK_SIZE_K": 32, "GROUP_SIZE_M": 8},num_stages=1,num_warps=4),
            triton.Config({"BLOCK_SIZE_M": 32, "BLOCK_SIZE_N": 32, "BLOCK_SIZE_K": 32, "GROUP_SIZE_M": 8},num_stages=1,num_warps=4),
            triton.Config({"BLOCK_SIZE_M": 32, "BLOCK_SIZE_N": 32, "BLOCK_SIZE_K": 32, "GROUP_SIZE_M": 8},num_stages=4,num_warps=4),
            triton.Config({"BLOCK_SIZE_M": 128, "BLOCK_SIZE_N": 128, "BLOCK_SIZE_K": 32, "GROUP_SIZE_M": 8},num_stages=4,num_warps=4),
            triton.Config({"BLOCK_SIZE_M": 64, "BLOCK_SIZE_N": 128, "BLOCK_SIZE_K": 32, "GROUP_SIZE_M": 8},num_stages=4,num_warps=4),
            triton.Config({"BLOCK_SIZE_M": 128, "BLOCK_SIZE_N": 32, "BLOCK_SIZE_K": 32, "GROUP_SIZE_M": 8},num_stages=4,num_warps=4),
            triton.Config({"BLOCK_SIZE_M": 64, "BLOCK_SIZE_N": 64, "BLOCK_SIZE_K": 32, "GROUP_SIZE_M": 8},num_stages=4,num_warps=4),
        ],
        key=["M", "N", "K"],
        nearest_power_of_two=True,
        prune_configs_by={
            "early_config_prune": custom_autotune.kernel_config_pruner,
            "perf_model": None,
            "top_k": None,
        },
    )
@triton.jit
def _ternary_mm_kernel(
        # Pointers to matrices
        a_ptr, b_ptr, c_ptr,
        # Matrix dimensions
        M, N, K,
        n_bits,
        # The stride variables represent how much to increase the ptr by when moving by 1
        # element in a particular dimension. E.g. `stride_am` is how much to increase `a_ptr`
        # by to get the element one row down (A has M rows).
        stride_am, stride_ak,
        stride_bk, stride_bn,
        stride_cm, stride_cn,
        # Kernel parameters
        BLOCK_SIZE_M: tl.constexpr, BLOCK_SIZE_N: tl.constexpr, BLOCK_SIZE_K: tl.constexpr,
        GROUP_SIZE_M: tl.constexpr,
        ACTIVATION: tl.constexpr,
):
    """Kernel for computing the matmul C = A x B.
        A has shape (M, K), int8
        B has shape (K//n_bits, N), int8, packed
        C has shape (M, N),
        """

    pid = tl.program_id(axis=0)
    num_pid_m = tl.cdiv(M, BLOCK_SIZE_M)
    num_pid_n = tl.cdiv(N, BLOCK_SIZE_N)
    num_pid_k = tl.cdiv(K, BLOCK_SIZE_K)
    num_pid_in_group = GROUP_SIZE_M * num_pid_n
    group_id = pid // num_pid_in_group
    first_pid_m = group_id * GROUP_SIZE_M
    group_size_m = min(num_pid_m - first_pid_m, GROUP_SIZE_M)
    pid_m = first_pid_m + (pid % group_size_m)
    pid_n = (pid % num_pid_in_group) // group_size_m
    # Create pointers
    offs_am = (pid_m * BLOCK_SIZE_M + tl.arange(0, BLOCK_SIZE_M)) % M
    offs_bn = (pid_n * BLOCK_SIZE_N + tl.arange(0, BLOCK_SIZE_N)) % N
    offs_k = tl.arange(0, BLOCK_SIZE_K)
    a_ptrs = a_ptr + (offs_am[:, None] * stride_am + offs_k[None, :] * stride_ak)  # (BLOCK_SIZE_M, BLOCK_SIZE_K)
    a_mask = offs_am[:, None] < M
    b_ptrs = b_ptr + (
                (offs_k[:, None] // n_bits) * stride_bk + offs_bn[None, :] * stride_bn)  # (BLOCK_SIZE_K, BLOCK_SIZE_N)

    c_dtype = tl.load(c_ptr + stride_cm).dtype # here we load the first element of c to see its dtype
    # shifter is used to extract each 2 bit of each element in the int matrix
    shifter = (offs_k % n_bits)[:, None] * 2

    accumulator = tl.zeros((BLOCK_SIZE_M, BLOCK_SIZE_N), dtype=tl.float32)
    for k in range(0, num_pid_k):
        a = (tl.load(a_ptrs, mask=a_mask, other=0.0))
        b = tl.load(b_ptrs)
        # We extract the 2 bits of each element in the int matrix
        b = (b >> shifter) & 0x3
        # We need to map back the value 2 -> -1
        b = tl.where(b == 0x2, -1, b)

        b = b.to(a.dtype)  # To be sure a.dtype == b_values.dtype

        accumulator += tl.dot(a, b, out_dtype=c_dtype)
        a_ptrs += BLOCK_SIZE_K
        b_ptrs += (BLOCK_SIZE_K // n_bits) * stride_bk

    c_ptrs = c_ptr + stride_cm * offs_am[:, None] + stride_cn * offs_bn[None, :]
    c_mask = (offs_am[:, None] < M) & (offs_bn[None, :] < N)
    tl.store(c_ptrs, accumulator, mask=c_mask)


def bitmat_(a, b, int_per_2_bits=4, activation="", out_dtype=torch.float16):
    """
        a: int8 tensor (..., K)
        b: int8 packed tensor (K // int_per_2_bit, N)
        c: float16 tensor (..., N)
        n_bits: int, number of bits that each element in b represents
    """
    # Check constraints.
    assert a.shape[-1] == b.shape[-2] * int_per_2_bits, "Incompatible dimensions"
    assert a.is_contiguous(), "A must be contiguous"
    assert b.is_contiguous(), "B must be contiguous"
    assert int_per_2_bits in [4, 8, 16, 32], "n_bits must be 4, 8, 16, 32"

    x = a.view(-1, a.shape[-1]) # flatten the tensor

    M, K = x.shape
    _, N = b.shape


    # Allocates output.
    c = torch.empty((M, N), device=a.device, dtype=out_dtype).contiguous()
    # 1D launch kernel where each block gets its own program.
    grid = lambda META: (
        triton.cdiv(M, META['BLOCK_SIZE_M']) * triton.cdiv(N, META['BLOCK_SIZE_N']),
    )

    # print(f"Launching kernel with M = {M}, N = {N}, K = {K}, n_bits = {n_bits}, activation = {activation}")

    _ternary_mm_kernel[grid](
        x, b, c,
        M, N, K,
        int_per_2_bits,
        x.stride(0), x.stride(1),
        b.stride(0), b.stride(1),
        c.stride(0), c.stride(1),
        ACTIVATION=activation
    )
    # bring c back to the original shape
    c = c.view(a.shape[:-1] + (N,))
    return c


import torch

from .utils.bitmat import bitmat
from .utils.rmsnorm import RMSLayerNorm
from .utils.packing import pack_ternary, unpack_ternary
from .utils.bitmat import terniarize

class BitLinear(torch.nn.Module):
    """
    A linear layer that uses packed terniary matrix multiplication.
    """

    def __init__(self, in_features, out_features, bias=None, eps=1e-5, keep_rms_in_32b=False, dtype=torch.float16):
        super(BitLinear, self).__init__()
        self.eps = eps
        self.in_features = in_features
        self.out_features = out_features
        self.dtype = dtype
        self.register_buffer('weight', torch.zeros((out_features, in_features), dtype=torch.int8))
        self.scale_w = torch.nn.Parameter(torch.Tensor(1))
        if bias:
            self.bias = torch.nn.Parameter(torch.Tensor(out_features))
        else:
            self.register_parameter('bias', None)
        self.norm = RMSLayerNorm(out_features, eps) # added this in favor of models like gemma that needs hp normalization
        self.keep_rms_in_32b = keep_rms_in_32b
        self._post_init()


    """
    def a .to() to keep eps precision
    """


    def _post_init(self):
        #crea un var dei parametri del modello così da poter inizializzare i pesi e i bias
        # Inizializza i pesi utilizzando l'inizializzazione di Kaiming
        params = torch.nn.Parameter(torch.zeros((self.out_features, self.in_features), dtype=self.dtype))
        torch.nn.init.kaiming_normal_(params, mode='fan_out', nonlinearity='relu')
        terniarized_val, self.scale_w.data = terniarize(params)
        del params
        self.register_buffer('weight',pack_ternary(terniarized_val))

        if self.bias is not None:
            torch.nn.init.constant_(self.bias, 0)

    def convert_weights_to_parameters(self):
        # Converti i pesi in torch.nn.Parameter di tipo float16 per il training.
        if self.weight.dtype == torch.int8:
            unpacked_weight = unpack_ternary(self.weight)
            half_weight = (unpacked_weight / self.scale_w).to(self.dtype)
            self.weight = torch.nn.Parameter(half_weight)
            self.scale_w = None# <- this is done so that the bitmat kernel knows we're training

    def convert_weights_to_packed(self):
        # Converti i pesi indietro in PackedParameter di tipo int8 dopo il training.
        if isinstance(self.weight, torch.nn.Parameter):
            terniarized_weight, scale_weight = terniarize(self.weight.data)
            packed_weights = pack_ternary(terniarized_weight)
            self.scale_w = torch.nn.Parameter(scale_weight)
            del self.weight # <- this is done so that torch doesn't trow an error when trying to convert the nn.Parameter to PackedParameter
            self.register_buffer('weight', packed_weights)

    def train(self, mode=True):
        super().train(mode)
        device = next(self.parameters()).device
        if mode:
            self.convert_weights_to_parameters()
        else:
            self.convert_weights_to_packed()
        return self.to(device)

    def forward(self, x):
        if self.training and (self.weight.dtype == torch.int8):
            # Just to make sure the weights are in the right format even if the user forgot to call train()
            self.convert_weights_to_parameters()
        x_dtype = x.dtype
        x = self.norm(x.to(self.norm.weight.dtype)).to(x_dtype)
        output = bitmat(self.weight.data, x, scale_w=self.scale_w)
        if self.bias is not None:
            output += self.bias.unsqueeze(0).expand_as(output)
        return output



from typing import Tuple

import torch

from ..triton_kernels.bitmat_kernel import bitmat_
from .packing import pack_ternary


def terniarize(weights: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Terniarizes the weights and returns the scale.
    """
    dtype = weights.dtype
    scale = 1 / torch.max(weights.abs().mean(), torch.tensor(1e-5))
    return torch.clamp((weights * scale).to(torch.int8), -1, 1), scale.to(dtype)

def quantize_activations(x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Quantizes the activations and returns the scale for each row.
    """
    dtype = x.dtype
    scale = (127 / torch.max(x.abs().max(dim=-1).values, torch.tensor(1e-5))).unsqueeze(-1)
    return torch.clamp((x * scale), -127, 128).to(torch.int8), scale.to(dtype)


class BitMat(torch.autograd.Function):
    @staticmethod
    @torch.cuda.amp.custom_fwd
    def forward(ctx, W, X, scale_w=None):
        """
        During the forward pass, we ternarize the weights, pack them and then quantize the activations.
        We then perform the bit matrix multiplication and return the scaled results.
        ternarization:
        scale_w = 1 / mean(abs(W))                              | STE
        W = clip(round(W * scale_w), -1, 1)                     | STE
        packing:
        packed_w = 4 int8 -> 1 int8                             | STE
        quantization:
        scale_x = 127 / max(abs(X))                             | STE
        X = clip(round(X * scale_x), -127, 128)                 | STE
        bit matrix multiplication:
        Y = X @ w_packed.t()                                    | dot product
        Y = Y / scale_w / scale_x)                              | STE
        """
        if scale_w is None:
            dtype = W.dtype
            W, scale_w = terniarize(W)
            #packed_w = pack_ternary(W, 4) -> this is actually not efficent atm
            ctx.save_for_backward(X)
            X, scale_x = quantize_activations(X)
            y = X.to(dtype) @ W.to(dtype).t()
            #y = batched_bitmat(X, packed_w) -> this is actually not efficent atm
            return y / scale_w / scale_x
        else:
            X, scale_x = quantize_activations(X)
            y = bitmat_(X, W.t().contiguous())
            return y / scale_w / scale_x


    @staticmethod
    @torch.cuda.amp.custom_bwd
    def backward(ctx, grad_output):
        X = ctx.saved_tensors[0]
        # get dW
        grad_W =  (grad_output.transpose(1,2) @ X).mean(dim=0)
        return grad_W, None, None

def bitmat(W: torch.Tensor, X: torch.Tensor, scale_w) -> torch.Tensor:
    return BitMat.apply(W, X, scale_w)


import torch
from torch import nn
from tqdm import tqdm
from ..bitlinear import BitLinear

from transformers import AutoModel, GemmaConfig, MistralConfig, LlamaConfig

# Importing custom hijack classes for specific models
from .model_hijacks.gemma_1_58b import Gemma158ForCausalLM
from .model_hijacks.mistral_1_58b import Mistral158ForCausalLM
from .model_hijacks.llama_1_58b import Llama158ForCausalLM


def convert_hf_model(model: AutoModel) -> AutoModel:
    """
    Convert a Hugging Face model to use BitLinear layers instead of Linear layers.
    """
    model_config = model.config
    del model

    # initialize a tqdm progress bar
    pbar = tqdm(total=1, desc="Converting model to 1.58Bit")
    # Check the type of the model and initialize the corresponding hijacked model
    if isinstance(model_config, GemmaConfig):
        hijacked_model = Gemma158ForCausalLM(model_config)
    elif isinstance(model_config, MistralConfig):
        hijacked_model = Mistral158ForCausalLM(model_config)
    elif isinstance(model_config, LlamaConfig):
        hijacked_model = Llama158ForCausalLM(model_config)
    else:
        raise RuntimeError("Unsupported model type. Please open an issue on GitHub citing the model you are using")

    pbar.update(1)
    return hijacked_model


def apply_bitlinear_to_hf_model(model: AutoModel, parent_name='') -> AutoModel:
    """
    Recursively replace Linear layers with BitLinear layers in the model,
    except for layers within 'lm_head'.
    """
    # Store any RMS layer configurations encountered
    rms_layers = {}

    # Remove RMS layers and store their configurations
    for name, module in list(model.named_children()):
        if 'RMS' in type(module).__name__:
            if hasattr(module, 'eps'):
                rms_eps = module.eps
            elif hasattr(module, 'epsilon'):
                rms_eps = module.epsilon
            elif hasattr(module, 'variance_epsilon'):
                rms_eps = module.variance_epsilon
            else:
                raise RuntimeError(
                    "Model type not mappable, please open an issue on GitHub citing the model you are using")

            # save weights and eps
            rms_layers[name] = {'eps': rms_eps}

            delattr(model, name)
    # Replace Linear layers with BitLinear layers
    for name, module in model.named_children():
        full_name = f'{parent_name}.{name}' if parent_name else name

        if 'lm_head' not in full_name and isinstance(module, nn.Linear):
            # Replace Linear layer with BitLinear
            bit_linear = BitLinear(module.in_features, module.out_features,
                                   rms_layers.get(f"{name}_rms", {}).get('eps', torch.tensor(1e-5)),
                                   module.bias is not None)
            setattr(model, name, bit_linear)
        else:
            # Recursively apply to child modules
            apply_bitlinear_to_hf_model(module, full_name)

    return model



import os
import json

from .model_hijacks import *  # noqa: F401


class Auto158ModelForCausalLM:
    @classmethod
    def load_model_conf(cls, path):
        if not os.path.exists(os.path.join(path, 'config.json')):
            raise ValueError(
                f"Cannot find a config file (config.json), at the specified directory please check the path")
        with open(os.path.join(path, 'config.json'), 'r') as f:
            model_conf = json.load(f)
            return model_conf['architectures'][0]

    @classmethod
    def from_pretrained(cls, *args, **kwargs):
        model_class_name = cls.load_model_conf(args[0])
        # Utilizza il mapping per ottenere il nome della classe del modello
        if not model_class_name:
            raise ValueError(
                f"The model {args[0]} was not found, this mean it has not been mapped yet, please open an issue on the github repository")

        # Ottieni la classe del modello utilizzando il suo nome
        model_class = globals()[model_class_name]

        # Restituisci l'istanza del modello
        return model_class.from_pretrained(*args, **kwargs)

    @classmethod
    def from_config(cls, *args, **kwargs):
        model_class_name = cls.load_model_conf(args[0])
        # Utilizza il mapping per ottenere il nome della classe del modello
        if not model_class_name:
            raise ValueError(
                f"The model {args[0]} was not found, this mean it has not been mapped yet, please open an issue on the github repository")

        # Ottieni la classe del modello utilizzando il suo nome
        model_class = globals()[model_class_name]

        # Restituisci l'istanza del modello
        return model_class.from_config(*args, **kwargs)



from bitmat.bitlinear import BitLinear
from bitmat.utils.bitmat import terniarize
from bitmat.utils.packing import pack_ternary


def pack_ternary_model(model, n_element_in_one_int=4):
    """
    Pack every BitLinear layer in the model.
    """
    for name, module in model.named_children():
        if isinstance(module, BitLinear):
            module.convert_weights_to_packed()

        else:
            pack_ternary_model(module, n_element_in_one_int=n_element_in_one_int)
    return model



__all__ = ['BitLinear', 'convert_hf_model', 'pack_ternary', 'unpack_ternary', 'Gemma158ForCausalLM', 'Mistral158ForCausalLM', 'Llama158ForCausalLM']
from .utils.convert_hf_model import convert_hf_model
from .utils.packing import pack_ternary, unpack_ternary
from .bitlinear import BitLinear
from .utils.model_hijacks.gemma_1_58b import Gemma158ForCausalLM
from .utils.model_hijacks.mistral_1_58b import Mistral158ForCausalLM
from .utils.model_hijacks.llama_1_58b import Llama158ForCausalLM


from setuptools import setup, find_packages

setup(
    name='bitmat-tl',
    version='0.3.1',
    author='Marco Lironi',
    author_email='marcolironi@astramind.ai',
    description='An efficent implementation for the paper: "The Era of 1-bit LLMs"',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/astramind-ai/BitMat/tree/main',
    packages=find_packages(),
    install_requires=[
        'torch',
        'triton',
        'transformers',
        'bitsandbytes',
        'numpy',
        'tqdm'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



import torch
from ..triton_kernels.rmsnorm_kernel import fast_rms_layernorm


class RMSLayerNorm(torch.nn.Module):
    def __init__(self, shape, eps=1e-5):
        super(RMSLayerNorm, self).__init__()
        self.eps = eps
        self.weight = torch.nn.Parameter(torch.randn(shape))

    def forward(self, x):
        return fast_rms_layernorm(self.weight, x, self.eps).to(x.dtype)
